import Eris from 'eris';
import { AwaitingObject, AwaitMessageOptions, ErisaOptions, Formatter, Matchable, MiddlewareHandler } from './types';
declare type MiddlewareEvent = Matchable | Matchable[];
declare type MiddlewareFunction = MiddlewareHandler | MiddlewareHandler[] | void;
/**
 * The main Erisa client.
 */
export default class Erisa extends Eris.Client {
    /** Collection containing the functions used to handle events. */
    handlers: Map<Matchable, MiddlewareHandler[]>;
    /** Collection containing objects for messages that are currently being waited for. */
    currentlyAwaiting: Map<string, AwaitingObject>;
    /** A regular object intended to store extensions from middleware, such as a command container. */
    extensions: {
        [x: string]: any;
    };
    constructor(token: string, options?: ErisaOptions);
    /**
     * Registers global middleware to the client.
     *
     * @param handlers An array of functions to run on any event.
     * @returns The current client instance.
     */
    use(...handlers: MiddlewareFunction[]): this;
    /**
     * Registers middleware for specific events to the client.
     *
     * @param events The events to apply the handlers to.
     * @param handlers An array of functions to run for the provided events.
     * @returns The current client instance.
     */
    use(events: MiddlewareEvent, ...handlers: MiddlewareFunction[]): this;
    /**
     * Registers pairs of middleware functions and event names. Intended for modules to utilise.
     *
     * @param pairs Array of pairs of middleware functions and events, or just functions.
     * @returns The current client instance.
     */
    use(pairs: Array<[MiddlewareEvent, MiddlewareFunction] | [MiddlewareFunction]> | void): this;
    /**
     * Remove middleware from events.
     *
     * @param events The events to remove the middleware from.
     * @param handlers The specific middleware to remove.
     * @returns The current client instance.
     */
    disuse(events: Matchable | Matchable[], ...handlers: Array<MiddlewareHandler | MiddlewareHandler[]>): this;
    disuse(...handlers: Array<MiddlewareHandler | MiddlewareHandler[]>): this;
    awaitMessage(channelID: string, userID: string, options?: AwaitMessageOptions): Promise<Eris.Message>;
    /**
     * Emits the provided event with the provided arguments, if any.
     * Also emits a wildcard "*" event for internal use.
     *
     * @param event Event to emit.
     * @param args Arguments for the event.
     * @returns `true` if the event had listeners, `false` otherwise.
     */
    emit(event: string | symbol, ...args: any[]): boolean;
    /**
     * Creates a function applied to `client.on` in order to prevent overload of event listeners.
     *
     * @param ev Event to create a handler function for.
     * @returns Handler function.
     */
    handleEvent(ev: string): (...args: any[]) => Promise<void>;
    /**
     * Formats a provided Eris object into a string form, since they don't have `.toString` methods.
     *
     * @param obj Object to format.
     * @param alt Whether to enable alternate behaviour. Disables showing discriminator if formatting users.
     * @param formatter Custom formatter function. Intended for use for non-Eris objects.
     * @returns The formatted string.
     */
    format<T>(obj: T, { alt, formatter }?: {
        alt?: boolean;
        formatter?: Formatter<T>;
    }): string;
}
export {};
