"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const eris_1 = __importDefault(require("eris"));
const minimatch_1 = __importDefault(require("minimatch"));
const awaitMessageHandler_1 = __importDefault(require("./awaitMessageHandler"));
const util_1 = require("./util");
const types_1 = require("./types");
/**
 * The main Erisa client.
 */
class Erisa extends eris_1.default.Client {
    constructor(token, options = {}) {
        super(token, options.erisOptions);
        /** Collection containing the functions used to handle events. */
        this.handlers = new Map();
        /** Collection containing objects for messages that are currently being waited for. */
        this.currentlyAwaiting = new Map();
        /** A regular object intended to store extensions from middleware, such as a command container. */
        this.extensions = {};
        this.on('*', this.handleEvent('*'));
        this.use('messageCreate', awaitMessageHandler_1.default);
    }
    use(...args) {
        const flattenedArgs = util_1.flat(args).filter(v => v);
        const setHandlers = (ev, handlers) => {
            if (!this.handlers.get(ev))
                this.handlers.set(ev, handlers);
            else
                this.handlers.set(ev, this.handlers.get(ev).concat([], handlers)); // typescript is dumb here :(
            if (typeof ev === 'string' && !this.eventNames().includes(ev))
                this.on(ev, this.handleEvent(ev));
        };
        // Handle multi-registration (passing an array of pairs to #use).
        if (args.length === 1 &&
            Array.isArray(args[0]) &&
            Array.isArray(args[0][0]))
            for (const pair of args[0])
                this.use(...pair);
        // Handle only functions passed (register to *).
        else if (typeof args[0] === 'function' ||
            flattenedArgs.reduce((m, v) => m && typeof v === 'function', true))
            setHandlers('*', flattenedArgs);
        else {
            let [events, ...handlers] = args;
            handlers = util_1.flat(handlers).filter(v => v); // Clean out void functions.
            events = !Array.isArray(events) ? [events] : events;
            for (const ev of events)
                setHandlers(ev, handlers);
        }
        return this;
    }
    disuse(...args) {
        const flattenedArgs = util_1.flat(args);
        const removeHandlers = (ev, handlers_) => {
            if (!this.handlers.get(ev) && ev !== '*')
                return;
            const handlers = handlers_.length ? handlers_ : this.handlers.get(ev);
            for (const handler of handlers) {
                const ourHandlers = ev === '*'
                    ? Array.from(this.handlers).filter(([, hndlrs]) => hndlrs.includes(handler))
                    : [[ev, this.handlers.get(ev)]]; // tslint:disable-line
                for (const [event, handlers__] of ourHandlers) {
                    const copied = handlers__.slice(); // Work on a new object, instead of operating on the reference.
                    copied.splice(copied.indexOf(handler), 1);
                    // Clean up as we go, especially when removing `*`.
                    if (!copied.length)
                        this.handlers.delete(event);
                    else
                        this.handlers.set(event, copied);
                }
            }
            if (this.handlers.get(ev) && !this.handlers.get(ev).length) {
                if (typeof ev === 'string')
                    this.removeAllListeners(ev);
                this.handlers.delete(ev);
            }
        };
        if (typeof args[0] === 'function' ||
            flattenedArgs.reduce((m, v) => m && typeof v === 'function', true))
            removeHandlers('*', flattenedArgs);
        else {
            let [events, ...handlers] = args;
            handlers = util_1.flat(handlers);
            events = typeof events === 'string' ? [events] : events;
            for (const ev of events)
                removeHandlers(ev, handlers);
        }
        return this;
    }
    awaitMessage(channelID, userID, options) {
        let resolve, reject;
        const id = channelID + userID;
        const { timeout = 15000, filter = () => true } = options || {};
        const promise = new Promise((...args) => {
            [resolve, reject] = args;
        });
        // Composes an object that resembles a semi-deconstructed promise, so that it can be returned now, and resolved/rejected at a later time.
        const deferred = {
            promise,
            resolve,
            reject
        };
        const timer = setTimeout(() => {
            deferred.reject(new types_1.AwaitTimeout('Message await expired.'));
            this.currentlyAwaiting.delete(id);
        }, timeout);
        this.currentlyAwaiting.set(id, {
            p: deferred,
            filter,
            timer
        });
        return deferred.promise;
    }
    /**
     * Emits the provided event with the provided arguments, if any.
     * Also emits a wildcard "*" event for internal use.
     *
     * @param event Event to emit.
     * @param args Arguments for the event.
     * @returns `true` if the event had listeners, `false` otherwise.
     */
    emit(event, ...args) {
        super.emit('*', event, ...args);
        return super.emit(event, ...args);
    }
    /**
     * Creates a function applied to `client.on` in order to prevent overload of event listeners.
     *
     * @param ev Event to create a handler function for.
     * @returns Handler function.
     */
    handleEvent(ev) {
        if (ev === '*')
            return function (event, ...args) {
                return __awaiter(this, void 0, void 0, function* () {
                    const matchingEvents = Array.from(this.handlers.keys()).filter(k => (k instanceof RegExp ? k.test(event) : minimatch_1.default(event, k)) &&
                        k !== event);
                    const handlers = util_1.flat(Array.from(this.handlers.entries())
                        .filter(([event]) => matchingEvents.includes(event))
                        .map(([, handler]) => handler));
                    for (const handler of handlers) // eslint-disable-next-line no-await-in-loop
                        yield handler({ event, erisa: this }, ...args);
                });
            }.bind(this);
        return function (...args) {
            return __awaiter(this, void 0, void 0, function* () {
                if (!this.handlers.get(ev))
                    return;
                for (const handler of this.handlers.get(ev)) // eslint-disable-next-line no-await-in-loop
                    yield handler({ event: ev, erisa: this }, ...args);
            });
        }.bind(this);
    }
    /**
     * Formats a provided Eris object into a string form, since they don't have `.toString` methods.
     *
     * @param obj Object to format.
     * @param alt Whether to enable alternate behaviour. Disables showing discriminator if formatting users.
     * @param formatter Custom formatter function. Intended for use for non-Eris objects.
     * @returns The formatted string.
     */
    format(obj, { alt, formatter } = {}) {
        // Should we store formatters in an internal map and let users add/remove as they please without passing them to this?
        if (formatter)
            return formatter(obj, alt);
        if (obj instanceof eris_1.default.Member)
            return `${obj.nick || obj.username}${!alt ? `#${obj.discriminator}` : ''}`;
        else if (obj instanceof eris_1.default.User)
            return `${obj.username}${!alt ? `#${obj.discriminator}` : ''}`;
        else if (obj instanceof eris_1.default.Role)
            return obj.mentionable ? obj.mention : obj.name;
        else if (obj instanceof eris_1.default.Channel)
            return obj.mention;
        else if (obj instanceof eris_1.default.Guild)
            return obj.name;
        else
            throw new util_1.FormatError(obj);
    }
}
exports.default = Erisa;
